module Bytewise
  module Ast
    #
    # This is something like a syntax tree.
    # It can hold other nodes like tags, chunks, etc.
    #
    class Body < Node
      attr_accessor :nodes, :blocks, :current_child, :context

      def initialize
        @nodes = []
        # Used while rendering, so that the node knows "where" it is.
        @current_child = nil
      end

      #
      # Adds a node or some chunk
      #
      # @param node String|Node
      #
      def insert(node, before: nil, after: nil)
        if node.is_a? Node
          if before.nil? && !after.nil?
            # Insert after
            @nodes.insert(@nodes.index(after) + 1, node)
          elsif after.nil? && !before.nil?
            # Insert before
            @nodes.insert(@nodes.index(before), node)
          else
            # Add to the end
            @nodes << node
          end

          node.parent = self
        elsif node.is_a? String
          add_chunk(node)
        elsif node.is_a? Array
          node.each { |n| insert n, before: before, after: after }
        end
      end

      alias_method :<<, :insert

      #
      # Adds chunk to the body
      #
      def add_chunk(content)
        content = content.to_s unless content.is_a?(String)

        unless content.empty?
          if content == "\n"
            @nodes << Bytewise::Ast::Chars::NewLine.new
          elsif @nodes.last.class == Bytewise::Ast::Chunk
            # Test if the last element was a chunk and append it
            @nodes.last << content
          else
            @nodes << Bytewise::Ast::Chunk.new(content)
          end
        end
      end


      def to_s
        string = ''

        @nodes.each do |node|
          string += node.to_s
        end
      end

      #
      # Called by the evaluator
      #
      def render(*args)
        result = ''

        @nodes.each_with_index do |node, index|
          @current_child = index
          result += node.render(*args).to_s
        end

        result
      end

      #
      # Imports the nodes of another body
      #
      # @param body ::Bytewise::Ast::Body
      #
      def import_body(body)
        body.nodes.each do |node|
          newnode = node.clone
          newnode.parent = self

          self << newnode
        end
      end

      #
      # Finds the last block with the given name
      #
      def find_blocks(name, exclude: nil)
        found = []

        @nodes.each do |node|
          if node.is_a? ::Bytewise::Ast::Body
            blocks = node.find_blocks(name, exclude: exclude)
            blocks.each { |b| found << b }
          elsif node.is_a? ::BraceMarkup::Ast::Tags::Block
            if node.name == name && node != exclude
              found << node
            end
          end
        end

        # Find in parent, if parent is a body
        if @parent.is_a? Body
          @parent.find_blocks(name, exclude: exclude).each { |n| found << n }
        end



        found
      end

      def children?
        @nodes.length > 0
      end

      def remove_child(child)
        @nodes.delete(child)
      end

      alias_method :'+=', :add_chunk
    end
  end
end